@@ -14,13 +14,14 @@ import (
 	"github.com/restic/restic/internal/backend"
 	"github.com/restic/restic/internal/backend/layout"
 	"github.com/restic/restic/internal/backend/location"
+	"github.com/restic/restic/internal/backend/util"
 	"github.com/restic/restic/internal/debug"
 	"github.com/restic/restic/internal/errors"
-	"github.com/restic/restic/internal/restic"
+	"github.com/restic/restic/internal/feature"
 )
 
-// make sure the rest backend implements restic.Backend
-var _ restic.Backend = &Backend{}
+// make sure the rest backend implements backend.Backend
+var _ backend.Backend = &Backend{}
 
 // Backend uses the REST protocol to access data stored on a server.
 type Backend struct {
@@ -30,6 +31,20 @@ type Backend struct {
 	layout.Layout
 }
 
+// restError is returned whenever the server returns a non-successful HTTP status.
+type restError struct {
+	backend.Handle
+	StatusCode int
+	Status     string
+}
+
+func (e *restError) Error() string {
+	if e.StatusCode == http.StatusNotFound && e.Handle.Type.String() != "invalid" {
+		return fmt.Sprintf("%v does not exist", e.Handle)
+	}
+	return fmt.Sprintf("unexpected HTTP response (%v): %v", e.StatusCode, e.Status)
+}
+
 func NewFactory() location.Factory {
 	return location.NewHTTPBackendFactory("rest", ParseConfig, StripPassword, Create, Open)
 }
@@ -58,6 +73,17 @@ func Open(_ context.Context, cfg Config, rt http.RoundTripper) (*Backend, error)
 	return be, nil
 }
 
+func drainAndClose(resp *http.Response) error {
+	_, err := io.Copy(io.Discard, resp.Body)
+	cerr := resp.Body.Close()
+
+	// return first error
+	if err != nil {
+		return errors.Errorf("drain: %w", err)
+	}
+	return cerr
+}
+
 // Create creates a new REST on server configured in config.
 func Create(ctx context.Context, cfg Config, rt http.RoundTripper) (*Backend, error) {
 	be, err := Open(ctx, cfg, rt)
@@ -65,7 +91,7 @@ func Create(ctx context.Context, cfg Config, rt http.RoundTripper) (*Backend, er
 		return nil, err
 	}
 
-	_, err = be.Stat(ctx, restic.Handle{Type: restic.ConfigFile})
+	_, err = be.Stat(ctx, backend.Handle{Type: backend.ConfigFile})
 	if err == nil {
 		return nil, errors.New("config file already exists")
 	}
@@ -80,18 +106,12 @@ func Create(ctx context.Context, cfg Config, rt http.RoundTripper) (*Backend, er
 		return nil, err
 	}
 
+	if err := drainAndClose(resp); err != nil {
+		return nil, err
+	}
+
 	if resp.StatusCode != http.StatusOK {
-		return nil, fmt.Errorf("server response unexpected: %v (%v)", resp.Status, resp.StatusCode)
-	}
-
-	_, err = io.Copy(io.Discard, resp.Body)
-	if err != nil {
-		return nil, err
-	}
-
-	err = resp.Body.Close()
-	if err != nil {
-		return nil, err
+		return nil, &restError{backend.Handle{}, resp.StatusCode, resp.Status}
 	}
 
 	return be, nil
@@ -101,11 +121,6 @@ func (b *Backend) Connections() uint {
 	return b.connections
 }
 
-// Location returns this backend's location (the server's URL).
-func (b *Backend) Location() string {
-	return b.url.String()
-}
-
 // Hasher may return a hash function for calculating a content hash for the backend
 func (b *Backend) Hasher() hash.Hash {
 	return nil
@@ -118,7 +133,7 @@ func (b *Backend) HasAtomicReplace() bool {
 }
 
 // Save stores data in the backend at the handle.
-func (b *Backend) Save(ctx context.Context, h restic.Handle, rd restic.RewindReader) error {
+func (b *Backend) Save(ctx context.Context, h backend.Handle, rd backend.RewindReader) error {
 	ctx, cancel := context.WithCancel(ctx)
 	defer cancel()
 
@@ -136,43 +151,45 @@ func (b *Backend) Save(ctx context.Context, h restic.Handle, rd restic.RewindRea
 	req.ContentLength = rd.Length()
 
 	resp, err := b.client.Do(req)
-
-	var cerr error
-	if resp != nil {
-		_, _ = io.Copy(io.Discard, resp.Body)
-		cerr = resp.Body.Close()
-	}
-
 	if err != nil {
 		return errors.WithStack(err)
 	}
 
-	if resp.StatusCode != http.StatusOK {
-		return errors.Errorf("server response unexpected: %v (%v)", resp.Status, resp.StatusCode)
+	if err := drainAndClose(resp); err != nil {
+		return err
 	}
 
-	return errors.Wrap(cerr, "Close")
-}
+	if resp.StatusCode != http.StatusOK {
+		return &restError{h, resp.StatusCode, resp.Status}
+	}
 
-// notExistError is returned whenever the requested file does not exist on the
-// server.
-type notExistError struct {
-	restic.Handle
-}
-
-func (e *notExistError) Error() string {
-	return fmt.Sprintf("%v does not exist", e.Handle)
+	return nil
 }
 
 // IsNotExist returns true if the error was caused by a non-existing file.
 func (b *Backend) IsNotExist(err error) bool {
-	var e *notExistError
-	return errors.As(err, &e)
+	var e *restError
+	return errors.As(err, &e) && e.StatusCode == http.StatusNotFound
+}
+
+func (b *Backend) IsPermanentError(err error) bool {
+	if b.IsNotExist(err) {
+		return true
+	}
+
+	var rerr *restError
+	if errors.As(err, &rerr) {
+		if rerr.StatusCode == http.StatusRequestedRangeNotSatisfiable || rerr.StatusCode == http.StatusUnauthorized || rerr.StatusCode == http.StatusForbidden {
+			return true
+		}
+	}
+
+	return false
 }
 
 // Load runs fn with a reader that yields the contents of the file at h at the
 // given offset.
-func (b *Backend) Load(ctx context.Context, h restic.Handle, length int, offset int64, fn func(rd io.Reader) error) error {
+func (b *Backend) Load(ctx context.Context, h backend.Handle, length int, offset int64, fn func(rd io.Reader) error) error {
 	r, err := b.openReader(ctx, h, length, offset)
 	if err != nil {
 		return err
@@ -201,7 +218,7 @@ func (b *Backend) Load(ctx context.Context, h restic.Handle, length int, offset
 	return err
 }
 
-func (b *Backend) openReader(ctx context.Context, h restic.Handle, length int, offset int64) (io.ReadCloser, error) {
+func (b *Backend) openReader(ctx context.Context, h backend.Handle, length int, offset int64) (io.ReadCloser, error) {
 	req, err := http.NewRequestWithContext(ctx, "GET", b.Filename(h), nil)
 	if err != nil {
 		return nil, errors.WithStack(err)
@@ -215,60 +232,48 @@ func (b *Backend) openReader(ctx context.Context, h restic.Handle, length int, o
 	req.Header.Set("Accept", ContentTypeV2)
 
 	resp, err := b.client.Do(req)
-
 	if err != nil {
-		if resp != nil {
-			_, _ = io.Copy(io.Discard, resp.Body)
-			_ = resp.Body.Close()
-		}
 		return nil, errors.Wrap(err, "client.Do")
 	}
 
-	if resp.StatusCode == http.StatusNotFound {
-		_ = resp.Body.Close()
-		return nil, &notExistError{h}
+	if resp.StatusCode != http.StatusOK && resp.StatusCode != http.StatusPartialContent {
+		_ = drainAndClose(resp)
+		return nil, &restError{h, resp.StatusCode, resp.Status}
 	}
 
-	if resp.StatusCode != http.StatusOK && resp.StatusCode != http.StatusPartialContent {
-		_ = resp.Body.Close()
-		return nil, errors.Errorf("unexpected HTTP response (%v): %v", resp.StatusCode, resp.Status)
+	if feature.Flag.Enabled(feature.BackendErrorRedesign) && length > 0 && resp.ContentLength != int64(length) {
+		return nil, &restError{h, http.StatusRequestedRangeNotSatisfiable, "partial out of bounds read"}
 	}
 
 	return resp.Body, nil
 }
 
 // Stat returns information about a blob.
-func (b *Backend) Stat(ctx context.Context, h restic.Handle) (restic.FileInfo, error) {
+func (b *Backend) Stat(ctx context.Context, h backend.Handle) (backend.FileInfo, error) {
 	req, err := http.NewRequestWithContext(ctx, http.MethodHead, b.Filename(h), nil)
 	if err != nil {
-		return restic.FileInfo{}, errors.WithStack(err)
+		return backend.FileInfo{}, errors.WithStack(err)
 	}
 	req.Header.Set("Accept", ContentTypeV2)
 
 	resp, err := b.client.Do(req)
 	if err != nil {
-		return restic.FileInfo{}, errors.WithStack(err)
+		return backend.FileInfo{}, errors.WithStack(err)
 	}
 
-	_, _ = io.Copy(io.Discard, resp.Body)
-	if err = resp.Body.Close(); err != nil {
-		return restic.FileInfo{}, errors.Wrap(err, "Close")
-	}
-
-	if resp.StatusCode == http.StatusNotFound {
-		_ = resp.Body.Close()
-		return restic.FileInfo{}, &notExistError{h}
+	if err = drainAndClose(resp); err != nil {
+		return backend.FileInfo{}, err
 	}
 
 	if resp.StatusCode != http.StatusOK {
-		return restic.FileInfo{}, errors.Errorf("unexpected HTTP response (%v): %v", resp.StatusCode, resp.Status)
+		return backend.FileInfo{}, &restError{h, resp.StatusCode, resp.Status}
 	}
 
 	if resp.ContentLength < 0 {
-		return restic.FileInfo{}, errors.New("negative content length")
+		return backend.FileInfo{}, errors.New("negative content length")
 	}
 
-	bi := restic.FileInfo{
+	bi := backend.FileInfo{
 		Size: resp.ContentLength,
 		Name: h.Name,
 	}
@@ -277,7 +282,7 @@ func (b *Backend) Stat(ctx context.Context, h restic.Handle) (restic.FileInfo, e
 }
 
 // Remove removes the blob with the given name and type.
-func (b *Backend) Remove(ctx context.Context, h restic.Handle) error {
+func (b *Backend) Remove(ctx context.Context, h backend.Handle) error {
 	req, err := http.NewRequestWithContext(ctx, "DELETE", b.Filename(h), nil)
 	if err != nil {
 		return errors.WithStack(err)
@@ -285,32 +290,25 @@ func (b *Backend) Remove(ctx context.Context, h restic.Handle) error {
 	req.Header.Set("Accept", ContentTypeV2)
 
 	resp, err := b.client.Do(req)
-
 	if err != nil {
 		return errors.Wrap(err, "client.Do")
 	}
 
-	if resp.StatusCode == http.StatusNotFound {
-		_ = resp.Body.Close()
-		return &notExistError{h}
+	if err = drainAndClose(resp); err != nil {
+		return err
 	}
 
 	if resp.StatusCode != http.StatusOK {
-		return errors.Errorf("blob not removed, server response: %v (%v)", resp.Status, resp.StatusCode)
+		return &restError{h, resp.StatusCode, resp.Status}
 	}
 
-	_, err = io.Copy(io.Discard, resp.Body)
-	if err != nil {
-		return errors.Wrap(err, "Copy")
-	}
-
-	return errors.Wrap(resp.Body.Close(), "Close")
+	return nil
 }
 
 // List runs fn for each file in the backend which has the type t. When an
 // error occurs (or fn returns an error), List stops and returns it.
-func (b *Backend) List(ctx context.Context, t restic.FileType, fn func(restic.FileInfo) error) error {
-	url := b.Dirname(restic.Handle{Type: t})
+func (b *Backend) List(ctx context.Context, t backend.FileType, fn func(backend.FileInfo) error) error {
+	url := b.Dirname(backend.Handle{Type: t})
 	if !strings.HasSuffix(url, "/") {
 		url += "/"
 	}
@@ -322,7 +320,6 @@ func (b *Backend) List(ctx context.Context, t restic.FileType, fn func(restic.Fi
 	req.Header.Set("Accept", ContentTypeV2)
 
 	resp, err := b.client.Do(req)
-
 	if err != nil {
 		return errors.Wrap(err, "List")
 	}
@@ -333,25 +330,31 @@ func (b *Backend) List(ctx context.Context, t restic.FileType, fn func(restic.Fi
 			// already ignores missing directories, but misuses "not found" to
 			// report certain internal errors, see
 			// https://github.com/rclone/rclone/pull/7550 for details.
-			return nil
+			return drainAndClose(resp)
 		}
 	}
 
 	if resp.StatusCode != http.StatusOK {
-		return errors.Errorf("List failed, server response: %v (%v)", resp.Status, resp.StatusCode)
+		_ = drainAndClose(resp)
+		return &restError{backend.Handle{Type: t}, resp.StatusCode, resp.Status}
 	}
 
 	if resp.Header.Get("Content-Type") == ContentTypeV2 {
-		return b.listv2(ctx, resp, fn)
+		err = b.listv2(ctx, resp, fn)
+	} else {
+		err = b.listv1(ctx, t, resp, fn)
 	}
 
-	return b.listv1(ctx, t, resp, fn)
+	if cerr := drainAndClose(resp); cerr != nil && err == nil {
+		err = cerr
+	}
+	return err
 }
 
 // listv1 uses the REST protocol v1, where a list HTTP request (e.g. `GET
 // /data/`) only returns the names of the files, so we need to issue an HTTP
 // HEAD request for each file.
-func (b *Backend) listv1(ctx context.Context, t restic.FileType, resp *http.Response, fn func(restic.FileInfo) error) error {
+func (b *Backend) listv1(ctx context.Context, t backend.FileType, resp *http.Response, fn func(backend.FileInfo) error) error {
 	debug.Log("parsing API v1 response")
 	dec := json.NewDecoder(resp.Body)
 	var list []string
@@ -360,7 +363,7 @@ func (b *Backend) listv1(ctx context.Context, t restic.FileType, resp *http.Resp
 	}
 
 	for _, m := range list {
-		fi, err := b.Stat(ctx, restic.Handle{Name: m, Type: t})
+		fi, err := b.Stat(ctx, backend.Handle{Name: m, Type: t})
 		if err != nil {
 			return err
 		}
@@ -385,7 +388,7 @@ func (b *Backend) listv1(ctx context.Context, t restic.FileType, resp *http.Resp
 
 // listv2 uses the REST protocol v2, where a list HTTP request (e.g. `GET
 // /data/`) returns the names and sizes of all files.
-func (b *Backend) listv2(ctx context.Context, resp *http.Response, fn func(restic.FileInfo) error) error {
+func (b *Backend) listv2(ctx context.Context, resp *http.Response, fn func(backend.FileInfo) error) error {
 	debug.Log("parsing API v2 response")
 	dec := json.NewDecoder(resp.Body)
 
@@ -402,7 +405,7 @@ func (b *Backend) listv2(ctx context.Context, resp *http.Response, fn func(resti
 			return ctx.Err()
 		}
 
-		fi := restic.FileInfo{
+		fi := backend.FileInfo{
 			Name: item.Name,
 			Size: item.Size,
 		}
@@ -429,5 +432,5 @@ func (b *Backend) Close() error {
 
 // Delete removes all data in the backend.
 func (b *Backend) Delete(ctx context.Context) error {
-	return backend.DefaultDelete(ctx, b)
+	return util.DefaultDelete(ctx, b)
 }
