@@ -3,6 +3,6 @@ package retry
 import "testing"
 
 // TestFastRetries reduces the initial retry delay to 1 millisecond
-func TestFastRetries(t testing.TB) {
+func TestFastRetries(_ testing.TB) {
 	fastRetries = true
 }
