@@ -5,6 +5,7 @@ package restorer
 
 import (
 	"context"
+	"io/fs"
 	"os"
 	"path/filepath"
 	"syscall"
@@ -12,12 +13,11 @@ import (
 	"time"
 
 	"github.com/restic/restic/internal/repository"
-	"github.com/restic/restic/internal/restic"
 	rtest "github.com/restic/restic/internal/test"
 	restoreui "github.com/restic/restic/internal/ui/restore"
 )
 
-func TestRestorerRestoreEmptyHardlinkedFileds(t *testing.T) {
+func TestRestorerRestoreEmptyHardlinkedFields(t *testing.T) {
 	repo := repository.TestRepository(t)
 
 	sn, _ := saveSnapshot(t, repo, Snapshot{
@@ -29,13 +29,9 @@ func TestRestorerRestoreEmptyHardlinkedFileds(t *testing.T) {
 				},
 			},
 		},
-	})
+	}, noopGetGenericAttributes)
 
-	res := NewRestorer(repo, sn, false, nil)
-
-	res.SelectFilter = func(item string, dstpath string, node *restic.Node) (selectedForRestore bool, childMayBeSelected bool) {
-		return true, true
-	}
+	res := NewRestorer(repo, sn, Options{})
 
 	tempdir := rtest.TempDir(t)
 	ctx, cancel := context.WithCancel(context.Background())
@@ -69,20 +65,15 @@ func getBlockCount(t *testing.T, filename string) int64 {
 	return st.Blocks
 }
 
-type printerMock struct {
-	filesFinished, filesTotal, allBytesWritten, allBytesTotal uint64
-}
-
-func (p *printerMock) Update(_, _, _, _ uint64, _ time.Duration) {
-}
-func (p *printerMock) Finish(filesFinished, filesTotal, allBytesWritten, allBytesTotal uint64, _ time.Duration) {
-	p.filesFinished = filesFinished
-	p.filesTotal = filesTotal
-	p.allBytesWritten = allBytesWritten
-	p.allBytesTotal = allBytesTotal
-}
-
 func TestRestorerProgressBar(t *testing.T) {
+	testRestorerProgressBar(t, false)
+}
+
+func TestRestorerProgressBarDryRun(t *testing.T) {
+	testRestorerProgressBar(t, true)
+}
+
+func testRestorerProgressBar(t *testing.T, dryRun bool) {
 	repo := repository.TestRepository(t)
 
 	sn, _ := saveSnapshot(t, repo, Snapshot{
@@ -95,14 +86,11 @@ func TestRestorerProgressBar(t *testing.T) {
 			},
 			"file2": File{Links: 1, Inode: 2, Data: "example"},
 		},
-	})
+	}, noopGetGenericAttributes)
 
 	mock := &printerMock{}
 	progress := restoreui.NewProgress(mock, 0)
-	res := NewRestorer(repo, sn, false, progress)
-	res.SelectFilter = func(item string, dstpath string, node *restic.Node) (selectedForRestore bool, childMayBeSelected bool) {
-		return true, true
-	}
+	res := NewRestorer(repo, sn, Options{Progress: progress, DryRun: dryRun})
 
 	tempdir := rtest.TempDir(t)
 	ctx, cancel := context.WithCancel(context.Background())
@@ -112,12 +100,43 @@ func TestRestorerProgressBar(t *testing.T) {
 	rtest.OK(t, err)
 	progress.Finish()
 
-	const filesFinished = 4
-	const filesTotal = filesFinished
-	const allBytesWritten = 10
-	const allBytesTotal = allBytesWritten
-	rtest.Assert(t, mock.filesFinished == filesFinished, "filesFinished: expected %v, got %v", filesFinished, mock.filesFinished)
-	rtest.Assert(t, mock.filesTotal == filesTotal, "filesTotal: expected %v, got %v", filesTotal, mock.filesTotal)
-	rtest.Assert(t, mock.allBytesWritten == allBytesWritten, "allBytesWritten: expected %v, got %v", allBytesWritten, mock.allBytesWritten)
-	rtest.Assert(t, mock.allBytesTotal == allBytesTotal, "allBytesTotal: expected %v, got %v", allBytesTotal, mock.allBytesTotal)
+	rtest.Equals(t, restoreui.State{
+		FilesFinished:   4,
+		FilesTotal:      4,
+		FilesSkipped:    0,
+		AllBytesWritten: 10,
+		AllBytesTotal:   10,
+		AllBytesSkipped: 0,
+	}, mock.s)
+}
+
+func TestRestorePermissions(t *testing.T) {
+	snapshot := Snapshot{
+		Nodes: map[string]Node{
+			"foo": File{Data: "content: foo\n", Mode: 0o600, ModTime: time.Now()},
+		},
+	}
+
+	repo := repository.TestRepository(t)
+	tempdir := filepath.Join(rtest.TempDir(t), "target")
+	ctx, cancel := context.WithCancel(context.Background())
+	defer cancel()
+
+	sn, id := saveSnapshot(t, repo, snapshot, noopGetGenericAttributes)
+	t.Logf("snapshot saved as %v", id.Str())
+
+	res := NewRestorer(repo, sn, Options{})
+	rtest.OK(t, res.RestoreTo(ctx, tempdir))
+
+	for _, overwrite := range []OverwriteBehavior{OverwriteIfChanged, OverwriteAlways} {
+		// tamper with permissions
+		path := filepath.Join(tempdir, "foo")
+		rtest.OK(t, os.Chmod(path, 0o700))
+
+		res = NewRestorer(repo, sn, Options{Overwrite: overwrite})
+		rtest.OK(t, res.RestoreTo(ctx, tempdir))
+		fi, err := os.Stat(path)
+		rtest.OK(t, err)
+		rtest.Equals(t, fs.FileMode(0o600), fi.Mode().Perm(), "unexpected permissions")
+	}
 }
